@@ -4,11 +4,14 @@ require "execjs/runtime"
 module ExecJS
   class ExternalRuntime < Runtime
     class Context < Runtime::Context
-      def initialize(runtime, source = "")
+      def initialize(runtime, source = "", options = {})
         source = encode(source)
 
         @runtime = runtime
         @source  = source
+
+        # Test compile context source
+        exec("")
       end
 
       def eval(source, options = {})
@@ -21,12 +24,19 @@ module ExecJS
 
       def exec(source, options = {})
         source = encode(source)
-        source = "#{@source}\n#{source}" if @source
+        source = "#{@source}\n#{source}" if @source != ""
         source = @runtime.compile_source(source)
 
         tmpfile = write_to_tempfile(source)
+
+        if ExecJS.cygwin?
+          filepath = `cygpath -m #{tmpfile.path}`.rstrip
+        else
+          filepath = tmpfile.path
+        end
+
         begin
-          extract_result(@runtime.exec_runtime(tmpfile.path))
+          extract_result(@runtime.exec_runtime(filepath), filepath)
         ensure
           File.unlink(tmpfile)
         end
@@ -54,14 +64,25 @@ module ExecJS
           tmpfile
         end
 
-        def extract_result(output)
-          status, value = output.empty? ? [] : ::JSON.parse(output, create_additions: false)
+        def extract_result(output, filename)
+          status, value, stack = output.empty? ? [] : ::JSON.parse(output, create_additions: false)
           if status == "ok"
             value
-          elsif value =~ /SyntaxError:/
-            raise RuntimeError, value
           else
-            raise ProgramError, value
+            stack ||= ""
+            real_filename = File.realpath(filename)
+            stack = stack.split("\n").map do |line|
+              line.sub(" at ", "")
+                  .sub(real_filename, "(execjs)")
+                  .sub(filename, "(execjs)")
+                  .strip
+            end
+            stack.reject! { |line| ["eval code", "eval@[native code]"].include?(line) }
+            stack.shift unless stack[0].to_s.include?("(execjs)")
+            error_class = value =~ /SyntaxError:/ ? RuntimeError : ProgramError
+            error = error_class.new(value)
+            error.set_backtrace(stack + caller)
+            raise error
           end
         end
     end
@@ -99,20 +120,25 @@ module ExecJS
         @binary ||= which(@command)
       end
 
-      def locate_executable(cmd)
-        if ExecJS.windows? && File.extname(cmd) == ""
-          cmd << ".exe"
+      def locate_executable(command)
+        commands = Array(command)
+        if ExecJS.windows? && File.extname(command) == ""
+          ENV['PATHEXT'].split(File::PATH_SEPARATOR).each { |p|
+            commands << (command + p)
+          }
         end
 
-        if File.executable? cmd
-          cmd
-        else
-          path = ENV['PATH'].split(File::PATH_SEPARATOR).find { |p|
-            full_path = File.join(p, cmd)
-            File.executable?(full_path) && File.file?(full_path)
-          }
-          path && File.expand_path(cmd, path)
-        end
+        commands.find { |cmd|
+          if File.executable? cmd
+            cmd
+          else
+            path = ENV['PATH'].split(File::PATH_SEPARATOR).find { |p|
+              full_path = File.join(p, cmd)
+              File.executable?(full_path) && File.file?(full_path)
+            }
+            path && File.expand_path(cmd, path)
+          end
+        }
       end
 
     protected
@@ -141,20 +167,71 @@ module ExecJS
         end
       end
 
-      def exec_runtime(filename)
-        io = IO.popen(binary.split(' ') << filename, @popen_options.merge({err: [:child, :out]}))
-        output = io.read
-        io.close
+      if ExecJS.windows?
+        def exec_runtime(filename)
+          path = Dir::Tmpname.create(['execjs', 'json']) {}
+          begin
+            command = binary.split(" ") << filename
+            `#{shell_escape(*command)} 2>&1 > #{path}`
+            output = File.open(path, 'rb', @popen_options) { |f| f.read }
+          ensure
+            File.unlink(path) if path
+          end
 
-        if $?.success?
-          output
-        else
-          raise RuntimeError, output
+          if $?.success?
+            output
+          else
+            raise exec_runtime_error(output)
+          end
+        end
+
+        def shell_escape(*args)
+          # see http://technet.microsoft.com/en-us/library/cc723564.aspx#XSLTsection123121120120
+          args.map { |arg|
+            arg = %Q("#{arg.gsub('"','""')}") if arg.match(/[&|()<>^ "]/)
+            arg
+          }.join(" ")
+        end
+      elsif RUBY_ENGINE == 'jruby'
+        require 'shellwords'
+
+        def exec_runtime(filename)
+          command = "#{Shellwords.join(binary.split(' ') << filename)} 2>&1"
+          io = IO.popen(command, @popen_options)
+          output = io.read
+          io.close
+
+          if $?.success?
+            output
+          else
+            raise exec_runtime_error(output)
+          end
+        end
+      else
+        def exec_runtime(filename)
+          io = IO.popen(binary.split(' ') << filename, @popen_options.merge({err: [:child, :out]}))
+          output = io.read
+          io.close
+
+          if $?.success?
+            output
+          else
+            raise exec_runtime_error(output)
+          end
         end
       end
       # Internally exposed for Context.
       public :exec_runtime
 
+      def exec_runtime_error(output)
+        error = RuntimeError.new(output)
+        lines = output.split("\n")
+        lineno = lines[0][/:(\d+)$/, 1] if lines[0]
+        lineno ||= 1
+        error.set_backtrace(["(execjs):#{lineno}"] + caller)
+        error
+      end
+
       def which(command)
         Array(command).find do |name|
           name, args = name.split(/\s+/, 2)
