@@ -1,5 +1,4 @@
-(function(program, execJS) { execJS(program) })(function() {
-  return eval(#{encode_source(source)});
+(function(program, execJS) { execJS(program) })(function() { #{source}
 }, function(program) {
   var output;
   try {
@@ -10,10 +9,10 @@
       try {
         print(JSON.stringify(['ok', result]));
       } catch (err) {
-        print('["err"]');
+        print(JSON.stringify(['err', '' + err, err.stack]));
       }
     }
   } catch (err) {
-    print(JSON.stringify(['err', '' + err]));
+    print(JSON.stringify(['err', '' + err, err.stack]));
   }
 });
